/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 19 janvier 2007
   derniere modification le 08/03/2007 */


#ifndef __SCEMESH_H__
#define __SCEMESH_H__

/*
#define SCE_MAX_ARRAYS (SCE_TEXCOORD7-SCE_POSITION+1)*/

/* NOTE: constante de merde :D */
#define SCE_MAX_BUFFERS 8


typedef struct sce_smeshvertexbuffer SCE_SMeshVertexBuffer;
struct sce_smeshvertexbuffer
{
    /* nombre de sommets, a utiliser avec SCE_CDrawBuffer */
    size_t count;
    
    SCEenum usage;              /* mode d'utilisation du tampon (GL_*_DRAW) */
    SCE_CVertexBuffer *buffer;  /* tampon */
    
    int builded;                /* booleen, true: le tampon est construit */
};


typedef struct sce_smeshindexbuffer SCE_SMeshIndexBuffer;
struct sce_smeshindexbuffer
{
    /* nombre d'indices, a utiliser avec SCE_CDrawIndexedBuffer */
    size_t count;
    
    SCEenum data_type;          /* type des donnees (GL_FLOAT, ...) */
    SCEenum usage;              /* mode d'utilisation du tampon (GL_*_DRAW) */
    SCE_CIndexBuffer *buffer;   /* tampon */
    
    int builded;                /* booleen, true: le tampon est construit */
};


typedef struct sce_mesh_mesh SCE_SMesh;
struct sce_mesh_mesh
{
    /* nombre de tampons de sommets initialises */
    unsigned int n_vb;
    
    /* nombre de tampons actifs pour le rendu
       si ce nombre est egal a 0, alors le count par defaut du mesh
       peut etre reinitialise a zero 0 */
    unsigned int n_active_vb;
    
    /* etat des tampons de sommets (active/desactive) */
    int vb_actives[SCE_MAX_BUFFERS];
    
    /* tampons de sommets */
    SCE_SMeshVertexBuffer vb[SCE_MAX_BUFFERS];
    
    
    
    /* nombre de tampons d'indices initialises */
    unsigned int n_ib;
    
    /* ID du tampon d'indices actif */
    unsigned int ib_active;
    
    /* tampons d'indices */
    SCE_SMeshIndexBuffer ib[SCE_MAX_BUFFERS];
    
    
    /* nombre de sommets qu'utilisent les vertex buffers actifs.
       si on tente d'activer un vertex buffer, on compare son count
       a celui-ci, s'il est different, on ne peut pas l'activer. */
    size_t active_vcount;
    
    /* idem, mais pour les indices */
    size_t active_icount;
    
    /* type des polygones par defaut */
    SCEenum polygon_type;
    
    
    /* booleen.
       true: mesh construit;
       0:    mesh non construit  */
    int builded;
    
    
    /* booleen.
       true: utilisation des indices (utilisation de i_count);
       0:    dessin a la volee (utilisation de v_count) */
    int use_indices;
};



/* initialise le gestionnaire de meshs */
int SCE_Mesh_Init(void);


/* quitte le gestionnaire */
int SCE_Mesh_Quit(void);


/* cree et initialise un nouveau mesh */
SCE_SMesh* SCE_Mesh_Create(void);


/* supprime un mesh existant */
int SCE_Mesh_Delete(SCE_SMesh*);


/* rajoute un tampon de sommets au mesh specifie */
int SCE_Mesh_AddVB(SCE_SMesh*, SCEenum);
void SCE_Mesh_DeleteVB(SCE_SMesh*, unsigned int);

/* rajoute un tampon d'indices au mesh specifie */
int SCE_Mesh_AddIB(SCE_SMesh*, SCEenum, SCEenum);
void SCE_Mesh_DeleteIB(SCE_SMesh*, unsigned int);


/* defini l'etat du tampon (sommets) specifie (active/desactive) */
int SCE_Mesh_SetActiveVB(SCE_SMesh*, unsigned int, int);

/* defini l'etat du tampon (indices) specifie (active/desactive) */
int SCE_Mesh_SetActiveIB(SCE_SMesh*, unsigned int, int);


/* envoie un tableau de donnees d'un type specifique au mesh specifie */
int SCE_Mesh_SetVertices(
  SCE_SMesh*,
  unsigned int,
  unsigned int,
  SCEenum,
  unsigned int,
  size_t,
  void*);


/* envoie un tableau de sommets entrelaces au mesh specifie */
/*int SCE_Mesh_Fill(
  SCE_SMesh*,
  unsigned int,
  SCE_CVertexDeclaration*,
  size_t,
  unsigned int,
  void*);*/


/* attribue un tableau d'indices au mesh specifie */
int SCE_Mesh_SetIndices(SCE_SMesh*, unsigned int, size_t, void*);


/* fonction de calcul des tangentes et des binormales pour un triangle donne.
   vertex et texcoord sont des tableaux dans lesquels la fonction se deplace
   librement selon les indices stockes dans index.
   la fonction ne depasse jamais index[2] */
void SCE_Mesh_ComputeTriangleTBN(
  float*, float*, unsigned int*,    /* vertex, texcoord, index */
  float*, float*, float*);          /* tangents, binormals, normals */


/* fonction qui genere les tangentes, binormales et normales pour des vertices
   et coordonnees de texture donnes. un tableau d'indices peut etre fourni
   en option, s'il n'y en a pas, les sommets seront prits l'un apres l'autre.
   la fonction a egalement besoin de connaitre le nombre de sommet que contient
   vertex et texcoord, et le nombre d'indices disponibles dans indices.
   le type des polygones est egalement requi, si celui-ci n'est pas supporte,
   une erreur est generee et la fonction renvoie SCE_ERROR apres avoir logge
   une erreur dans le gestionnaire d'erreur. */
int SCE_Mesh_ComputeTBN(
  SCEenum,
  float*, float*, unsigned int*,    /* vertex, texcoord, index */
  unsigned int, unsigned int,       /* icount, vcount */
  float*, float*, float*);          /* tangents, binormals, normals */


/* genere les vecteur tangents, binormals et normals d'un mesh donne.
   l'avzntage a utiliser cette fonction est qu'elle prend en compte le type
   des variables des tableaux de sommets du mesh. */
int SCE_Mesh_GenerateTBN(
  SCE_SMesh*,
  unsigned int, unsigned int, int,  /* ids des tampons: vertex texcoord index */
  float*, float*, float*,           /* tangents, binormals, normals */
  unsigned int);                    /* unite de texture */


/* construit le mesh specifie
   prepare le rendu
   on ne peut constuire qu'une seule fois un mesh
   par la suite si on souhaite le modifier,
   il faudra passer par la fonction Update */
int SCE_Mesh_Build(SCE_SMesh*);


/* defini le type mode de rendu du mesh (type de polygone) */
void SCE_Mesh_SetRenderMode(SCE_SMesh*, SCEenum);


/* met a jour les donnees du tampon specifie */
int SCE_Mesh_Update(SCE_SMesh*, unsigned int);


/* stock et defini le mesh specifie comme mesh actif pour le rendu */
void SCE_Mesh_Bind(SCE_SMesh*);

/* effectue le rendu du mesh specifie par Mesh_Bind */
void SCE_Mesh_Draw(SCEenum);



/* rendu d'une plage d'un tableau de meshs.
   la plage est definie tel qu'elle commence a first et fini a nb-1 */
void SCE_Mesh_Render(unsigned int, unsigned int, SCE_SMesh**);


#endif /* guard */
