/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 10 janvier 2007
   derniere modification le 08/03/2007 */



#include <SCE/SCEMinimal.h>

#include <string.h> /* pas la peine de le prendre de utils/... a moins que ? */

#include <SCE/core/SCECBuffers.h>


/* constantes internes */
#define SCE_BUFFER_OFFSET(p) ((char*)NULL + (p))

/* nombre d'attributs maximum... :/ (NOTE: cf plus bas) */
#define SCE_MAX_ATTRIBS 16


SCE_CVertexBuffer* SCE_CCreateVertexBuffer(void)
{
    SCE_CVertexBuffer *vb = NULL;
    
    vb = SCE_malloc(sizeof *vb);
    if(vb == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    vb->id = 0;
    vb->n_data = 0;
    vb->datas = NULL;
    vb->size = 0;
    vb->n_decs = 0;
    vb->dec = NULL;
    
    glGenBuffers(1, &vb->id);
    
    return (vb);
}

SCE_CIndexBuffer* SCE_CCreateIndexBuffer(void)
{
    SCE_CIndexBuffer *ib = NULL;
    
    ib = SCE_malloc(sizeof *ib);
    if(ib == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    ib->id = 0;
    ib->size = 0;
    ib->datas = NULL;
    ib->n_data = 0;
    
    glGenBuffers(1, &ib->id);
    
    return (ib);
}


int
SCE_CSetVertexBufferData(
  SCE_CVertexBuffer *vb,    /* tampon */
  size_t data_size,         /* taille des donnees */
  void *data)               /* donnees */
{
    if(vb != NULL)
    {
        /* allocation d'une structure de donnees supplementaire */
        vb->datas = SCE_realloc(vb->datas, (vb->n_data+1)*sizeof(*vb->datas));
        if(vb->datas == NULL)
        {
            Logger_LogSrc();
            return (SCE_ERROR);
        }
        
        /* affectation */
        vb->datas[vb->n_data].data_size = data_size;
        vb->datas[vb->n_data].data = data;
        vb->datas[vb->n_data].first = vb->size;
        
        /* par addition cumulative, pour empiler les donnes dans le buffer */
        vb->size += data_size;
        
        /* incrementation du compteur de nombre de donnees */
        vb->n_data++;
    }
    
    return (0);
}

int
SCE_CSetIndexBufferData(
  SCE_CIndexBuffer *ib,
  size_t data_size,
  void *data)
{
    if(ib != NULL)
    {
        /* allocation d'une structure de donnees supplementaire */
        ib->datas = SCE_realloc(ib->datas, (ib->n_data+1)*sizeof(*ib->datas));
        if(ib->datas == NULL)
        {
            Logger_LogSrc();
            return (SCE_ERROR);
        }
        
        /* affectation */
        ib->datas[ib->n_data].data_size = data_size;
        ib->datas[ib->n_data].data = data;
        ib->datas[ib->n_data].first = ib->size;
        
        /* par addition cumulative, pour empiler les donnes dans le buffer */
        ib->size += data_size;
        
        /* incrementation du compteur de nombre de donnees */
        ib->n_data++;
    }
    
    return (0);
}


size_t SCE_CGetVertexBufferFirst(SCE_CVertexBuffer *vb)
{
    if(vb != NULL)
    {
        return (vb->size);
    }
    
    return (0);
}


void
SCE_CSetVertexDeclaration(
  SCE_CVertexBuffer *vb,
  unsigned int n_decs,
  SCE_CVertexDeclaration *dec)
{
    if(vb != NULL && dec != NULL)
    {
        vb->dec = dec;
        vb->n_decs = n_decs;
    }
}


void SCE_CBuildVertexBuffer(SCE_CVertexBuffer *vb, SCEenum usage)
{
    unsigned int i;
    
    if(vb != NULL)
    {
        glBindBuffer(GL_ARRAY_BUFFER, vb->id);
        glBufferData(GL_ARRAY_BUFFER, vb->size, NULL, usage);
        
        for(i=0; i<vb->n_data; i++)
        {
            glBufferSubData(GL_ARRAY_BUFFER,
                            vb->datas[i].first,
                            vb->datas[i].data_size,
                            vb->datas[i].data);
        }
        
        glBindBuffer(GL_ARRAY_BUFFER, 0);
    }
}


void SCE_CBuildIndexBuffer(SCE_CIndexBuffer *ib, SCEenum usage)
{
    unsigned int i;
    
    if(ib != NULL)
    {
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ib->id);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, ib->size, NULL, usage);
        
        for(i=0; i<ib->n_data; i++)
        {
            glBufferSubData(GL_ELEMENT_ARRAY_BUFFER,
                            ib->datas[i].first,
                            ib->datas[i].data_size,
                            ib->datas[i].data);
        }
        
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
}


void* SCE_CLockVertexBuffer(SCE_CVertexBuffer *vb, SCEenum mode)
{
    if(vb != NULL)
    {
        glBindBuffer(GL_ARRAY_BUFFER, vb->id);
        return (glMapBuffer(GL_ARRAY_BUFFER, mode));
    }
    
    return (NULL);
}



void SCE_CUnlockVertexBuffer(SCE_CVertexBuffer *vb)
{
    if(vb != NULL)
    {
        glBindBuffer(GL_ARRAY_BUFFER, vb->id);
        glUnmapBuffer(GL_ARRAY_BUFFER);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
    }
}


void* SCE_CLockIndexBuffer(SCE_CIndexBuffer *ib, SCEenum mode)
{
    if(ib != NULL)
    {
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ib->id);
        return (glMapBuffer(GL_ELEMENT_ARRAY_BUFFER, mode));
    }
    
    return (NULL);
}

void SCE_CUnlockIndexBuffer(SCE_CIndexBuffer *ib)
{
    if(ib != NULL)
    {
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ib->id);
        glUnmapBuffer(GL_ELEMENT_ARRAY_BUFFER);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
}


int SCE_CUpdateVertexBuffer(SCE_CVertexBuffer *vb)
{
    unsigned int i;
    void *data = NULL;
    
    if(vb != NULL)
    {
        glBindBuffer(GL_ARRAY_BUFFER, vb->id);
        data = glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
        if(data == NULL)
        {
            Logger_Log(-1);
            Logger_LogMsg("OpenGL can't lock vertex buffer!");
            return (SCE_ERROR);
        }
        else
        {
            for(i=0; i<vb->n_data; i++)
            {
                memcpy(data + vb->datas[i].first,
                       vb->datas[i].data,
                       vb->datas[i].data_size);
            }
        }
        glUnmapBuffer(GL_ARRAY_BUFFER);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
    }
    
    return (SCE_OK);
}


int SCE_CUpdateIndexBuffer(SCE_CIndexBuffer *ib)
{
    unsigned int i;
    void *data;
    
    if(ib != NULL)
    {
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ib->id);
        data = glMapBuffer(GL_ELEMENT_ARRAY_BUFFER, GL_WRITE_ONLY);
        if(data == NULL)
        {
            Logger_Log(-1);
            Logger_LogMsg("OpenGL can't lock index buffer!");
            return (SCE_ERROR);
        }
        else
        {
            for(i=0; i<ib->n_data; i++)
            {
                memcpy(data + ib->datas[i].first,
                       ib->datas[i].data,
                       ib->datas[i].data_size);
            }
        }
        glUnmapBuffer(GL_ELEMENT_ARRAY_BUFFER);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
    
    return (SCE_OK);
}


void SCE_CUseVertexBuffer(SCE_CVertexBuffer *vb)
{
    static unsigned int n_attribs = 0;
    
    /* index des attributs de sommet
       sur ma machine, la limite d'attributs est a 16
       il faut voir ou est le maximum, et l'appliquer ici */
    static unsigned int attrib_index[SCE_MAX_ATTRIBS];
    
    unsigned int i;
    
    if(vb != NULL)
    {
        glBindBuffer(GL_ARRAY_BUFFER, vb->id);
        
        for(i=0; i<vb->n_decs; i++)
        {
            switch(vb->dec[i].v_type)
            {
                case SCE_POSITION:
                    glEnableClientState(GL_VERTEX_ARRAY);
                    glVertexPointer(vb->dec[i].n_vars,
                                    vb->dec[i].type,
                                    vb->dec[i].stride,
                                    SCE_BUFFER_OFFSET(vb->dec[i].first));
                    break;
                
                case SCE_NORMAL:
                    glEnableClientState(GL_NORMAL_ARRAY);
                    glNormalPointer(vb->dec[i].type,
                                    vb->dec[i].stride,
                                    SCE_BUFFER_OFFSET(vb->dec[i].first));
                    break;
                
                case SCE_COLOR:
                    glEnableClientState(GL_COLOR_ARRAY);
                    glColorPointer(vb->dec[i].n_vars,
                                   vb->dec[i].type,
                                   vb->dec[i].stride,
                                   SCE_BUFFER_OFFSET(vb->dec[i].first));
                    break;
                
                default:
                    if(vb->dec[i].v_type <= SCE_TEXCOORD7 &&
                       vb->dec[i].v_type >= SCE_TEXCOORD0)
                    {
                        /* coordonnees de texture */
                        glClientActiveTexture(
                            GL_TEXTURE0 + (vb->dec[i].v_type-SCE_TEXCOORD0) );
                        
                        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
                        glTexCoordPointer(vb->dec[i].n_vars,
                                          vb->dec[i].type,
                                          vb->dec[i].stride,
                                          SCE_BUFFER_OFFSET(vb->dec[i].first));
                    }
                    else
                    {
                        /* attribut */
                        
                        /* affectation de l'id de l'attribut
                           au tableau statique */
                        attrib_index[n_attribs] =
                            vb->dec[i].v_type - SCE_ATTRIB0;
                         
                         
                        glEnableVertexAttribArray(attrib_index[n_attribs]);
                        glVertexAttribPointer(attrib_index[n_attribs],
                                           vb->dec[i].n_vars,
                                           vb->dec[i].type,
                                           GL_FALSE,     /* ne pas normaliser */
                                           vb->dec[i].stride,
                                           SCE_BUFFER_OFFSET(vb->dec[i].first));
                        
                        /* + un attribut */
                        n_attribs++;
                        
                        #ifdef __SCE_DEBUG__
                        /* NOTE: !!! */
                        if(n_attribs >= SCE_MAX_ATTRIBS)
                        {
                            Logger_Log(SCE_INVALID_SIZE);
                            Logger_LogMsg("TMP+DEBUG:"
                                          "you can't set an vertex attribute "
                                          "if his index is greater of %d",
                                          SCE_MAX_ATTRIBS);
                        }
                        #endif
                    }
                    break;
            }
        }
    }
    else
    {
        /* par precaution, on desactive tout */
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);
        glDisableClientState(GL_COLOR_ARRAY);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        for(i=0; i<n_attribs; i++)
        {
            glDisableVertexAttribArray(attrib_index[i]);
            attrib_index[i] = 0;
        }
        n_attribs = 0;
        glBindBuffer(GL_ARRAY_BUFFER, 0);
    }
}

void SCE_CUseIndexBuffer(SCE_CIndexBuffer *ib)
{
    if(ib != NULL)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ib->id);
    else
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}


void SCE_CDrawBuffer(SCEenum p_type, GLint first, GLsizei count)
{
    glDrawArrays(p_type, first, count);
}

void SCE_CDrawIndexedBuffer(
  SCEenum p_type, size_t first, SCEenum i_type, GLsizei count)
{
    glDrawElements(p_type, count, i_type, SCE_BUFFER_OFFSET(first));
}


void SCE_CDeleteVertexBuffer(SCE_CVertexBuffer *vb)
{
    if(vb != NULL)
    {
        glDeleteBuffers(1, &vb->id);
        SCE_free(vb);
    }
}

void SCE_CDeleteIndexBuffer(SCE_CIndexBuffer *ib)
{
    if(ib != NULL)
    {
        glDeleteBuffers(1, &ib->id);
        SCE_free(ib);
    }
}
